#include <stdio.h>
#include <a.out.h>

/*
 * This program is intended for downloading program binaries to
 * a specially equipped prom monitor. It uses base 64 encoding, where
 * three bytes are sent as four characters. There are two record types,
 * those starting with '/' and those that do not. The records that start
 * with a '/' are used to perform special operations, all other records
 * contain 24 bits of data. The function send12 is used to send special
 * operations, whereas the function send24 is used to send data.
 * To further improve download performance, contiguous records of zero
 * are sent as a special record that includes a count value. The bss
 * section is cleared using these special records.
 */

#define DEBUG 0
#define BLKSIZE 200
#define swap(x,y)	((y==2)?byte_swap(x,y):word_swap(x,y))

/* special record types */
#define ZEROS "/Z"		/* record contains a count of zeros,
				   actually it's the number of 24-bit
				   records that are all zero */
#define BYTE "/B"		/* record contains a single byte */
#define CHKSUM "/C"		/* checksum to date */
#define ADDR "/A"		/* double length record containing a 32 bit
				   address */
#define END "/E"		/* end of download */

unsigned long addr;
int blksz = 0;
int chksum = 0;

/* define the other 2 chars needed for base 64 */
char exta = ',';
char extb = '.';

int filecount = 0;
int prom = 0;
int debug = 0;

main(argc,argv)
int argc;
char *argv[];
{
FILE *fp;
int i,j;

for (i=1;i<argc;i++) { /* first find all the options */
	if (argv[i][0] == '-') { /* must be an option */
		for (j=1;argv[i][j] != 0;j++) {
			if (argv[i][1] == 'p') prom = 1;
			else if (argv[i][1] == 'd') debug = 1;
			else printf("bad option %s ignored\n",argv[i]);
			}
		}
	else /* must be a file */ filecount++;
	}

if (filecount == 0) dofile(stdin,"stdin");

else {
	for (i=1;i<argc;i++) { /* do the files */
		if (argv[i][0] != '-') { /* must be a filename */
			fp = fopen(argv[i],"r");
			if (fp == 0) 
				fprintf(stderr,
				"can't open file (%s) ignored\n",argv[i]);
			else {
				dofile(fp,argv[i]);
				fclose(fp);
				}
			}
		}
	}
}

dofile(ifp,fname)
FILE *ifp;
char *fname;
{
int leflag;
unsigned long offset;
struct filehdr fhdr;
struct aouthdr ohdr;

fread(&fhdr,sizeof(fhdr),1,ifp);

if (fhdr.f_magic == 061001) { /* little endian */
	leflag = 1;
	swap(&fhdr.f_magic,sizeof(fhdr.f_magic));
	swap(&fhdr.f_nscns,sizeof(fhdr.f_nscns));
	swap(&fhdr.f_opthdr,sizeof(fhdr.f_opthdr));
	}
else leflag = 0;

if (debug) printf("magic=0%o nscns=%d ohdrsz=%d\n",
	fhdr.f_magic,fhdr.f_nscns,fhdr.f_opthdr);

if (fhdr.f_magic != 0540 && fhdr.f_magic != 0542) {
	fprintf(stderr,"0%o Bad magic number\n",fhdr.f_magic);
	exit(1);
	}

fread(&ohdr,sizeof(ohdr),1,ifp);

if (leflag) {
	swap(&ohdr.entry,sizeof(ohdr.entry));
	swap(&ohdr.tsize,sizeof(ohdr.tsize));
	swap(&ohdr.text_start,sizeof(ohdr.text_start));
	swap(&ohdr.dsize,sizeof(ohdr.dsize));
	swap(&ohdr.data_start,sizeof(ohdr.data_start));
	}

if (debug) printf("entry=%lx tsize=%lx tbase=%lx dsize=%lx dbase=%lx\n",
	ohdr.entry,ohdr.tsize,ohdr.text_start,ohdr.dsize,ohdr.data_start);

txaddr(ohdr.text_start);
offset = ((fhdr.f_nscns*sizeof(struct scnhdr))+sizeof(fhdr)+
	sizeof(ohdr)+15)&~15;
fseek(ifp,offset,0);
dosect(ifp,ohdr.tsize);
if (!prom) txaddr(ohdr.data_start);
dosect(ifp,ohdr.dsize);
if (!prom) clrbss(ohdr.bss_start,ohdr.bsize);
txeof(ohdr.entry);
}

dosect(infile,scnsize)
FILE *infile;
long scnsize;
{
int i,n,zcnt,reccnt;
unsigned char buf[3];
long cnt,bdat;

zcnt = 0;
reccnt = 0;
for (cnt=0;cnt+3<=scnsize;cnt+=3) {
	fread(buf,sizeof(buf[0]),3,infile); /* read 3 data bytes */
	bdat = (buf[0]<<16)+(buf[1]<<8)+buf[2]; /* convert to single word */
	chksum += bdat; /* update checksum */
	if (bdat == 0) { /* compress if the data is zero */
		zcnt++;
		if (zcnt < 4096) continue;
		zcnt--;
		}
	if (DEBUG) printf("%06x ",cnt);
	if (zcnt > 0) { /* send a special record for zeros */
		send12(ZEROS,zcnt);
		if (DEBUG) printf(" %03x\n",zcnt);
		reccnt++;
		zcnt = 0;
		}
	send24(bdat); /* send the data */
	reccnt++;
	if (DEBUG) printf(" %06x\n",bdat);
	}
if (zcnt > 0) { /* make sure we didn't end with zeros outstanding */
	send12(ZEROS,zcnt);
	if (DEBUG) printf(" %03x\n",zcnt);
	zcnt = 0;
	}
if (DEBUG) printf("cnt=%06x scnsize=%06x\n",cnt,scnsize);
for (i=0;i<scnsize-cnt;i++) { /* if sect size not divisable by 3 */
	fread(buf,sizeof(buf[0]),1,infile);
	send12(BYTE,buf[0]); /* send individual byte */
	if (DEBUG) printf(" %02x\n",buf[0]);
	}
if (DEBUG) printf("\n");
}

send12(type,val)
char *type;
int val;
{
char adat[4];

val &= 0xfff;
adat[0] = type[0];
adat[1] = type[1];
adat[2] = b2a(val>>6);
adat[3] = b2a(val);
send(adat,4);
}

send24(val)
unsigned int val;
{
char adat[4];
int i;

val &= 0xffffff;
for (i=3;i>=0;i--) {
	adat[i] = b2a(val);
	val>>=6;
	}
send(adat,4);
}

send(p,n)
char *p;
int n;
{

fwrite(p,sizeof(p[0]),n,stdout);
blksz += n;
if (blksz >= BLKSIZE) {
	putc('\n',stdout);
	blksz = 0;
	}
}

txchksum()
{

send12(CHKSUM,chksum);
chksum = 0;
}

txeof(addr)
unsigned long addr;
{

/* txchksum(chksum); */
txaddr(addr);
send12(END,0);
putc('\n',stdout);
}

txaddr(addr)
unsigned long addr;
{

send12(ADDR,addr>>24);
send24(addr);
if (DEBUG) printf(" %08x ",addr);
}

clrbss(addr,size)
long addr,size;
{
int i,rem;

txaddr(addr);
if (DEBUG) printf("%08x ",size);
rem = size % 3;
size /= 3;
while (size>4095) {
	send12(ZEROS,4095);
	size -= 4095;
	}
send12(ZEROS,size);
for (i=0;i<rem;i++) send12(BYTE,0);
}

a2b(c)
int c;
{
/* translate base64 (tx form) to binary
A-Z	0..25
a-z	26..51
0-9	52..61
exta	62
extb	63
*/

if (c >= 'A' && c <= 'Z') return(c-'A');
if (c >= 'a' && c <= 'z') return(c-'a'+26);
if (c >= '0' && c <= '9') return(c-'0'+52);
if (c == exta) return(62);
if (c == extb) return(63);
printf("a2a: Bad input value\n");
return(0);
}


b2a(c)
int c;
{
/* translate binary to base64 (tx form) 
0..25		A-Z
26..51		a-z
52..61		0-9
62
63
*/

c &= 0x3f;
if (c <= 25)  return('A'+c);
if (c <= 51) return('a'+c-26);
if (c <= 61) return('0'+c-52); 
if (c == 62) return(exta);
if (c == 63) return(extb);
printf("b2a: bad input value\n");
return(0);
}

word_swap(vp)
unsigned long *vp;
{
unsigned short v1,v2;

v1 = (*vp)>>16;
byte_swap(&v1);
v2 = (*vp)&0xffff;
byte_swap(&v2);
*vp = (v2<<16)+v1;
}

byte_swap(vp)
unsigned short *vp;
{

*vp = ((*vp)<<8)+((*vp)>>8);
}
